/*
	XDSLOAD.CPP
	2002, Lorenzo Lutti (lorenzo.lutti@tiscalinet.it)

	This program loads a process and handles its I/O by debugging it and trapping
	the PRIVILEGED_INSTRUCTION exception.

	Note 1: I've tested XDSLOAD only under Windows 2000; probably it runs under XP as well.
	The OpenThread() function is supported by Win2000, WinXP and WinME; I suppose
	there are other undocumented ways to get the same thing (retrieving a HANDLE from a TID) under
	the other OSes.

	Note 2: XDSLOAD works only for programs not designed for NT/2000/XP, that normally
	would crash	with the "Illegal instruction" debug message. It doesn't work with programs that
	access the I/O space in a "legal" way (kernel-mode driver, "unlocking" the I/O instructions
	by using Ke386SetIoAccessMap(), intercepting the I/O with a VDD and so on), simply because
	they don't throw any debugging exception when they access the I/O space.

	VERSION HISTORY

	08-oct-2002 (Lorenzo Lutti): first release.
*/


#include "stdafx.h"
#include "xdsload.h"
#include "xdsloadDlg.h"

#include "xds510\\xds510.h"

#define CHAR_LIMIT 10000000

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

typedef HANDLE (WINAPI *ptrOpenThread)(DWORD,BOOL,DWORD);
ptrOpenThread	OpenThread;

typedef struct
{
	CXdsloadDlg *dlg;
	PROCESS_INFORMATION pi;
} THREAD_DATA;

/////////////////////////////////////////////////////////////////////////////
// CXdsloadDlg dialog

CXdsloadDlg::CXdsloadDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CXdsloadDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CXdsloadDlg)
	//}}AFX_DATA_INIT
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);
	m_thread = NULL;
}

void CXdsloadDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CXdsloadDlg)
	DDX_Control(pDX, IDC_CHECK1, m_verbose);
	DDX_Control(pDX, IDC_RICHEDIT1, m_monitor);
	DDX_Control(pDX, IDOK, m_run);
	DDX_Control(pDX, IDC_EDIT1, m_path);
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CXdsloadDlg, CDialog)
	//{{AFX_MSG_MAP(CXdsloadDlg)
	ON_BN_CLICKED(IDC_BUTTON1, OnBrowse)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CXdsloadDlg message handlers

BOOL CXdsloadDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon

	// Retrieve OpenThread() address (not necessary if you have an updated PSDK)
	HMODULE hMod = LoadLibrary( "kernel32.dll");
	if( hMod == NULL)
	{
		AfxMessageBox( "LoadLibrary");
		EndDialog( IDCANCEL);
		return FALSE;
	}
	OpenThread = (ptrOpenThread)GetProcAddress( hMod, "OpenThread");
	if( OpenThread == NULL)
	{
		AfxMessageBox( "GetProcAddress");
		EndDialog( IDCANCEL);
		return FALSE;
	}

	// I/O grant
	m_wrapper.GiveIO();

	// XDS init
	XDS510_init();

	return TRUE;  // return TRUE  unless you set the focus to a control
}

void CXdsloadDlg::OnBrowse() 
{
	CFileDialog dlg( TRUE, NULL, NULL, OFN_HIDEREADONLY | OFN_PATHMUSTEXIST |
		OFN_FILEMUSTEXIST, "Executable files (*.exe)|*.exe||", this);

	if( dlg.DoModal() == IDOK)
		m_path.SetWindowText( dlg.GetPathName());
}

void CXdsloadDlg::OnOK() 
{
	m_run.EnableWindow( FALSE);

	m_monitor.SetWindowText( "");
	m_monitor.SetFocus();

	// Execute process (initially suspended)
	CString path;
	m_path.GetWindowText( path);
	PROCESS_INFORMATION pi;
	STARTUPINFO si;
	::FillMemory( &si, sizeof(si), 0);
	si.cb = sizeof( si);
	if( !CreateProcess( NULL, (char*)(LPCTSTR)path, NULL, NULL, FALSE, CREATE_SUSPENDED, NULL, NULL, &si, &pi))
	{
		m_run.EnableWindow( TRUE);
		AfxMessageBox( "CreateProcess");
		return;
	}

	// Create the debug thread
	THREAD_DATA *td = new THREAD_DATA;
	td->pi = pi;
	td->dlg = this;
	if( (m_thread = AfxBeginThread( DebugThread, (LPVOID)td  )) == NULL)
	{
		delete td;
		m_run.EnableWindow( TRUE);
		TerminateProcess( pi.hProcess, 0);
		AfxMessageBox( "CreateThread");
		return;		
	}
}

UINT CXdsloadDlg::DebugThread( LPVOID pParam)
{
	PROCESS_INFORMATION pi = ((THREAD_DATA*)pParam)->pi;
	CXdsloadDlg* dlg = ((THREAD_DATA*)pParam)->dlg;
	delete ((THREAD_DATA*)pParam);
	BOOL bVerbose = dlg->m_verbose.GetCheck() ? TRUE : FALSE;

	// Debug the process
	DEBUG_EVENT de;
	if( !DebugActiveProcess( pi.dwProcessId))
	{
		TerminateProcess( pi.hProcess, 0);
		AfxMessageBox( "DebugActiveProcess");
		dlg->m_run.EnableWindow( TRUE);
		return 0;
	}
	
	// Resume the process
	ResumeThread( pi.hThread);
	
	// Debugging loop
	while( TRUE)
	{
		// Wait indefinitely for a debug event
		if( !WaitForDebugEvent( &de, INFINITE))
		{
			TerminateProcess( pi.hProcess, 0);
			AfxMessageBox( "WaitForDebugEvent");
			dlg->m_run.EnableWindow( TRUE);
			return 0;
		}

		CString str;
		CString oldtxt;

		// Handle only the EXCEPTION_DEBUG_EVENT event
		if( de.dwDebugEventCode != EXCEPTION_DEBUG_EVENT)
		{
			// Event dump
			str = DebugEventString( de.dwDebugEventCode) + "\r\n";
			dlg->m_monitor.GetWindowText( oldtxt);
			int nChars = oldtxt.GetLength();
			if( nChars < CHAR_LIMIT)
			{
				dlg->m_monitor.SetSel( nChars, -1);
				dlg->m_monitor.ReplaceSel( str);				
			}
			
			// If the event is EXIT_PROCESS_DEBUG_EVENT, exit
			if( de.dwDebugEventCode == EXIT_PROCESS_DEBUG_EVENT)
			{
				dlg->m_run.EnableWindow( TRUE);
				return 0;				
			}

			// Always consider continuables these exceptions (TODO: probably wrong)
			ContinueDebugEvent( de.dwProcessId, de.dwThreadId, DBG_CONTINUE);
			continue;
		}
		else
		{
			// Handle only the STATUS_PRIVILEGED_INSTRUCTION exception
			if( de.u.Exception.ExceptionRecord.ExceptionCode != STATUS_PRIVILEGED_INSTRUCTION)
			{
				// Exception dump
				BOOL known;
				CString str = DebugEventString( de.dwDebugEventCode) +
					" " + ExceptionString( de.u.Exception.ExceptionRecord.ExceptionCode, &known) + "\r\n";
				dlg->m_monitor.GetWindowText( oldtxt);
				int nChars = oldtxt.GetLength();
				if( nChars < CHAR_LIMIT)
				{
					dlg->m_monitor.SetSel( nChars, -1);
					dlg->m_monitor.ReplaceSel( str);				
				}

				// Don't handle the unknown exceptions
				// (frequently the programs use them for their business)
				if( !known)
				{
					ContinueDebugEvent( de.dwProcessId, de.dwThreadId, DBG_EXCEPTION_NOT_HANDLED);
				}
				else
				// Continue for first chance (maybe someone can handle it...) or continuable exceptions
				if( de.u.Exception.dwFirstChance ||
					!(de.u.Exception.ExceptionRecord.ExceptionFlags & EXCEPTION_NONCONTINUABLE))
				{
					ContinueDebugEvent( de.dwProcessId, de.dwThreadId, DBG_CONTINUE);
				}
				else
				// Otherwise, game over.
				{
					TerminateProcess( pi.hProcess, 0);
					AfxMessageBox( "Second-chance exception");
					dlg->m_run.EnableWindow( TRUE);
					return 0;
				}
				
				continue;				
			}

			// Read the opcode that caused the exception
			HANDLE hProcess = OpenProcess( PROCESS_ALL_ACCESS, FALSE, de.dwProcessId);
			if( hProcess == NULL)
			{
				TerminateProcess( pi.hProcess, 0);
				AfxMessageBox( "OpenProcess");
				dlg->m_run.EnableWindow( TRUE);
				return 0;
			}
			unsigned char opcodes[2];
			DWORD dwBytesRead;
			if( !ReadProcessMemory( hProcess, de.u.Exception.ExceptionRecord.ExceptionAddress,
				opcodes, 2, &dwBytesRead))
			{
				TerminateProcess( pi.hProcess, 0);
				AfxMessageBox( "ReadProcessMemory");
				dlg->m_run.EnableWindow( TRUE);
				return 0;
			}
			CloseHandle( hProcess);
			
			// Read the thread context
			HANDLE hThread = OpenThread( THREAD_ALL_ACCESS, FALSE, de.dwThreadId);
			if( hThread == NULL)
			{
				TerminateProcess( pi.hProcess, 0);
				AfxMessageBox( "OpenThread");
				dlg->m_run.EnableWindow( TRUE);
				return 0;
			}
			CONTEXT context;
			context.ContextFlags = CONTEXT_INTEGER | CONTEXT_CONTROL;
			if( !GetThreadContext( hThread, &context))
			{
				CloseHandle( hThread);
				TerminateProcess( pi.hProcess, 0);
				AfxMessageBox( "GetThreadContext");
				dlg->m_run.EnableWindow( TRUE);
				return 0;
			}

			/*
				I'm going to handle only these opcodes:
				EC				IN AL,DX
				EE				OUT DX,AL
				66 ED			IN AX,DX
				66 EF			OUT DX,AX

				There are some others, i.e. double word IN and OUT and some privileged opcodes
				other than IN/OUT, like PUSHF. You can find them on Internet.
			*/

			CString str;
			int ipincr = 0;
			if( opcodes[0] == 0x66 && opcodes[1] == 0xED)		// IN AX,DX
			{
				ipincr = 2;
				DWORD port = context.Edx & 0x0000FFFF;
				if( bVerbose)
					str.Format( "IN AX,0x%04X\r\n", port);

				// TODO
				if ((port >= XDS510(0x00) && port < XDS510(0x20)) ||
					(port >= XDS510(0x400) && port < XDS510(0x410)) ||
					port == XDS510(0x800))
				{
					context.Eax &= 0xFFFF0000;
					context.Eax |= (unsigned)XDS510_read( port - XDS510_BASE) & 0x0000FFFF;
				}
				else
				{
					TRACE( "Direct I/O read attempted from port %x\n", port);
					context.Eax |= 0x0000FFFF;
				}
			}
			else if( opcodes[0] == 0x66 && opcodes[1] == 0xEF)	// OUT DX,AX
			{
				ipincr = 2;
				DWORD port = context.Edx & 0x0000FFFF;
				if( bVerbose)
					str.Format( "OUT 0x%04X,0x%04X\r\n", port, context.Eax & 0x0000FFFF);

				// TODO
				if ((port >= XDS510(0x00) && port < XDS510(0x20)) ||
					(port >= XDS510(0x400) && port < XDS510(0x410)) ||
					port == XDS510(0x800))
				{
					XDS510_write( port - XDS510_BASE, context.Eax & 0x0000FFFF);
				}
				else
				{ 
					TRACE( "Direct I/O write attempted to port %x\n", port);
				}
			}
/*			else if( opcodes[0] == 0xEC)						// IN AL,DX
			{
				ipincr = 1;
				if( bVerbose)
					str.Format( "IN AL,0x%04X\r\n", context.Edx & 0x0000FFFF);
				
				// TODO
			}
			else if( opcodes[0] == 0xEE)						// OUT DX,AL
			{
				ipincr = 1;
				if( bVerbose)
					str.Format( "OUT 0x%04X,0x%02X\r\n", context.Edx & 0x0000FFFF, context.Eax & 0x000000FF);
				
				// TODO
			}
*/
			if( ipincr > 0)
			{
				if( bVerbose)
				{
					dlg->m_monitor.GetWindowText( oldtxt);
					int nChars = oldtxt.GetLength();
					if( nChars < CHAR_LIMIT)
					{
						dlg->m_monitor.SetSel( nChars, -1);
						dlg->m_monitor.ReplaceSel( str);				
					}
				}

				// Advance EIP to the next instruction. This is enough to remove the exception
				context.Eip += ipincr;
				if( !SetThreadContext( hThread, &context))
				{
					CloseHandle( hThread);
					TerminateProcess( pi.hProcess, 0);
					AfxMessageBox( "SetThreadContext");
					dlg->m_run.EnableWindow( TRUE);
					return 0;
				}
				CloseHandle( hThread);
				ContinueDebugEvent( de.dwProcessId, de.dwThreadId, DBG_CONTINUE);
				continue;							
			}
			else	// Opcode not handled
			{
				CloseHandle( hThread);

				// Opcode dump
				str.Format( "%04X (edx=%08X eax=%08X)\r\n",
					opcodes[0] == 0x66 ? opcodes[0]<<8 | opcodes[1] : opcodes[0], context.Edx, context.Eax);
				dlg->m_monitor.GetWindowText( oldtxt);
				int nChars = oldtxt.GetLength();
				if( nChars < CHAR_LIMIT)
				{
					dlg->m_monitor.SetSel( nChars, -1);
					dlg->m_monitor.ReplaceSel( str);				
				}

				// Continue for first chances (maybe someone can handle it...) or continuable exceptions
				if( de.u.Exception.dwFirstChance ||
					!(de.u.Exception.ExceptionRecord.ExceptionFlags & EXCEPTION_NONCONTINUABLE))
				{
					ContinueDebugEvent( de.dwProcessId, de.dwThreadId, DBG_CONTINUE);
				}
				else
				// Otherwise, game over.
				{
					TerminateProcess( pi.hProcess, 0);
					AfxMessageBox( "Privileged instruction not handled, second-chance exception");
					dlg->m_run.EnableWindow( TRUE);
					return 0;
				}
				continue;				
			}
		}
	}
}

CString CXdsloadDlg::ExceptionString( DWORD code, BOOL* known) 
{
	if( known)
		*known = TRUE;

	switch( code)
	{
	case STATUS_WAIT_0                    :
		return CString( "WAIT_0");
	case STATUS_ABANDONED_WAIT_0          :
		return CString( "ABANDONED_WAIT_0");
	case STATUS_TIMEOUT                   :
		return CString( "TIMEOUT");
	case STATUS_PENDING                   :
		return CString( "PENDING");
	case STATUS_SEGMENT_NOTIFICATION      :
		return CString( "SEGMENT_NOTIFICATION");
	case STATUS_GUARD_PAGE_VIOLATION      :
		return CString( "GUARD_PAGE_VIOLATION");
	case STATUS_DATATYPE_MISALIGNMENT     :
		return CString( "DATATYPE_MISALIGNMENT");
	case STATUS_BREAKPOINT                :
		return CString( "BREAKPOINT");
	case STATUS_SINGLE_STEP               :
		return CString( "SINGLE_STEP");
	case STATUS_ACCESS_VIOLATION          :
		return CString( "ACCESS_VIOLATION");
	case STATUS_IN_PAGE_ERROR             :
		return CString( "IN_PAGE_ERROR");
	case STATUS_INVALID_HANDLE            :
		return CString( "INVALID_HANDLE");
	case STATUS_NO_MEMORY                 :
		return CString( "NO_MEMORY");
	case STATUS_ILLEGAL_INSTRUCTION       :
		return CString( "ILLEGAL_INSTRUCTION");
	case STATUS_NONCONTINUABLE_EXCEPTION  :
		return CString( "NONCONTINUABLE_EXCEPTION");
	case STATUS_INVALID_DISPOSITION       :
		return CString( "INVALID_DISPOSITION");
	case STATUS_ARRAY_BOUNDS_EXCEEDED     :
		return CString( "ARRAY_BOUNDS_EXCEEDED");
	case STATUS_FLOAT_DENORMAL_OPERAND    :
		return CString( "FLOAT_DENORMAL_OPERAND");
	case STATUS_FLOAT_DIVIDE_BY_ZERO      :
		return CString( "FLOAT_DIVIDE_BY_ZERO");
	case STATUS_FLOAT_INEXACT_RESULT      :
		return CString( "FLOAT_INEXACT_RESULT");
	case STATUS_FLOAT_INVALID_OPERATION   :
		return CString( "FLOAT_INVALID_OPERATION");
	case STATUS_FLOAT_OVERFLOW            :
		return CString( "FLOAT_OVERFLOW");
	case STATUS_FLOAT_STACK_CHECK         :
		return CString( "FLOAT_STACK_CHECK");
	case STATUS_FLOAT_UNDERFLOW           :
		return CString( "FLOAT_UNDERFLOW");
	case STATUS_INTEGER_DIVIDE_BY_ZERO    :
		return CString( "INTEGER_DIVIDE_BY_ZERO");
	case STATUS_INTEGER_OVERFLOW          :
		return CString( "INTEGER_OVERFLOW");
	case STATUS_PRIVILEGED_INSTRUCTION    :
		return CString( "PRIVILEGED_INSTRUCTION");
	case STATUS_STACK_OVERFLOW            :
		return CString( "STACK_OVERFLOW");
	case STATUS_CONTROL_C_EXIT            :
		return CString( "CONTROL_C_EXIT");
	case STATUS_FLOAT_MULTIPLE_FAULTS     :
		return CString( "FLOAT_MULTIPLE_FAULTS");
	case STATUS_FLOAT_MULTIPLE_TRAPS      :
		return CString( "FLOAT_MULTIPLE_TRAPS");
	case STATUS_ILLEGAL_VLM_REFERENCE     :
		return CString( "ILLEGAL_VLM_REFERENCE");
	default:
		if( known)
			*known = FALSE;
		CString str;
		str.Format( "Unknown (0x%08X)", code);
		return str;
	}
}

CString CXdsloadDlg::DebugEventString( DWORD code, BOOL *known)
{
	if( known)
		*known = TRUE;
	
	switch( code)
	{
	case EXCEPTION_DEBUG_EVENT:
		return CString( "EXCEPTION_DEBUG_EVENT");
	case CREATE_THREAD_DEBUG_EVENT:
		return CString( "CREATE_THREAD_DEBUG_EVENT");
	case CREATE_PROCESS_DEBUG_EVENT:
		return CString( "CREATE_PROCESS_DEBUG_EVENT");
	case EXIT_THREAD_DEBUG_EVENT:
		return CString( "EXIT_THREAD_DEBUG_EVENT");
	case EXIT_PROCESS_DEBUG_EVENT:
		return CString( "EXIT_PROCESS_DEBUG_EVENT");
	case LOAD_DLL_DEBUG_EVENT:
		return CString( "LOAD_DLL_DEBUG_EVENT");
	case UNLOAD_DLL_DEBUG_EVENT:
		return CString( "UNLOAD_DLL_DEBUG_EVENT");
	case OUTPUT_DEBUG_STRING_EVENT:
		return CString( "OUTPUT_DEBUG_STRING_EVENT");
	case RIP_EVENT:
		return CString( "RIP_EVENT");
	default:
		if( known)
			*known = FALSE;
		CString str;
		str.Format( "Unknown (0x%08X)", code);
		return str;	
	}
}

void CXdsloadDlg::OnCancel() 
{
	// TODO: Add extra cleanup here
	
	CDialog::OnCancel();
}
